/*	MVEPlayer.c - implements MVEPlayer function.
	Author: Shaktazuki
	with generous help from Mike Melanson, of FFMpeg and www.multimedia.cx fame
	Major goals:
	TODO: fix timer
	TODO: fix BG movie compatibility
	Minor goals:
	TODO: dragging the video output around messes up the file reads; pause while dragging or disable dragging (SDL_NOFRAME)
		or ... the callback can get its own data from the file using the file's audio frame indices (probably a bad idea).
	TODO: account for truecolor flag in opcode 0x05 
	TODO: change flag checking from CHKFLAG(FLAG) to just FLAG and make unsetting something like ~FLAG
	(may need to do that before this:)
	TODO: implement non-stereo audio decoder
	TODO: redefine mve data structs and code that uses them.
	(with respect to this last: the more I think about it, the original MVE stuff is nearly ideal).
*/
#include "MVEPlayer.h"
#include <assert.h>
#ifdef _MSC_VER
#include "sdl.h"
#else
#include <SDL.H>
#include <SDL/SDL.H>
#endif

/* define (potentially) handled opcode (opcode) types */
#define STOP_PLAYBACK			0x00
#define FETCH_NEXT_CHUNK		0x01
#define CREATE_TIMER			0x02
#define INIT_AUDIO_BUFFERS		0x03
#define START_AUDIO				0x04
#define INIT_VIDEO_BUFFERS		0x05
#define SEND_BUFFER_TO_DISPLAY	0x07
#define AUDIO_FRAME				0x08
#define SET_PALETTE				0x0C
#define SET_DECODING_MAP		0x0F
#define VIDEO_DATA				0x11

/**************/
/* flag masks */
/**************/

/* flow control flags */
#define QUITTING			0x00000001
#define PAUSED				0x00000002

/* timer flags */
#define TIMER_CREATED		0x00000004
#define TIMER_INIT			0x00000008

/* audio flags */
#define AUDIO_PLAYING		0x00000010
#define AUDIO_COMPRESSED	0x00000020
#define AUDIO_STEREO		0x00000040
#define SAMPLESIZE16		0x00000080

/* flags declaration */
Uint32 flags = 0;

/**************************/
/* internally used macros */
/**************************/

#define SETFLAG(x) flags |= x
#define NEGFLAG(x) flags &= ~x
#define CHKFLAG(x) (flags & x)
/* fetch LE words from a bytestream x */
#define LE16(x) (*(x) | ((*(x+1)) << 8))
/* write a signed word y to bytestream x */
#define WRITEWORD(x,y) {Sint16 * temp = (Sint16 *)x; *temp = y;}

/******************************/
/* internally used structures */
/******************************/

typedef struct CHUNK
{
	Uint16 length;
	Uint16 type;
} chunk;

typedef struct OPCODE
{
	Uint16 length;
	Uint8 type;
	Uint8 version;
} opcode;

typedef struct BUFSTRUCT
{
	Uint32 length;
	Uint8 *data;
} buffer;

/* Interplay DPCM table */
Sint16 interplay_delta_table[256] =
{
         0,      1,      2,      3,      4,      5,      6,      7,      8,      9,     10,     11,     12,     13,     14,     15,
        16,     17,     18,     19,     20,     21,     22,     23,     24,     25,     26,     27,     28,     29,     30,     31,
        32,     33,     34,     35,     36,     37,     38,     39,     40,     41,     42,     43,     47,     51,     56,     61,
        66,     72,     79,     86,     94,    102,    112,    122,    133,    145,    158,    173,    189,    206,    225,    245,
       267,    292,    318,    348,    379,    414,    452,    493,    538,    587,    640,    699,    763,    832,    908,    991,
      1081,   1180,   1288,   1405,   1534,   1673,   1826,   1993,   2175,   2373,   2590,   2826,   3084,   3365,   3672,   4008,
      4373,   4772,   5208,   5683,   6202,   6767,   7385,   8059,   8794,   9597,  10472,  11428,  12471,  13609,  14851,  16206,
     17685,  19298,  21060,  22981,  25078,  27367,  29864,  32589, -29973, -26728, -23186, -19322, -15105, -10503,  -5481,     -1,
         1,      1,   5481,  10503,  15105,  19322,  23186,  26728,  29973, -32589, -29864, -27367, -25078, -22981, -21060, -19298,
    -17685, -16206, -14851, -13609, -12471, -11428, -10472,  -9597,  -8794,  -8059,  -7385,  -6767,  -6202,  -5683,  -5208,  -4772,
     -4373,  -4008,  -3672,  -3365,  -3084,  -2826,  -2590,  -2373,  -2175,  -1993,  -1826,  -1673,  -1534,  -1405,  -1288,  -1180,
     -1081,   -991,   -908,   -832,   -763,   -699,   -640,   -587,   -538,   -493,   -452,   -414,   -379,   -348,   -318,   -292,
      -267,   -245,   -225,   -206,   -189,   -173,   -158,   -145,   -133,   -122,   -112,   -102,    -94,    -86,    -79,    -72,
       -66,    -61,    -56,    -51,    -47,    -43,    -42,    -41,    -40,    -39,    -38,    -37,    -36,    -35,    -34,    -33,
       -32,    -31,    -30,    -29,    -28,    -27,    -26,    -25,    -24,    -23,    -22,    -21,    -20,    -19,    -18,    -17,
       -16,    -15,    -14,    -13,    -12,    -11,    -10,     -9,     -8,     -7,     -6,     -5,     -4,     -3,     -2,     -1
};

/* screen buffers */
Uint32 *current_frame = NULL, *old_frame = NULL;

void MVEPlayerAudioCB(void *userdata, Uint8 *stream, Sint32 len);
void MVEPlayerDecodeAudio(buffer * input);
void MVEPlayerDecodeVideo(SDL_Surface * screen, buffer * video_data, buffer * decoding_map, Uint32 *palette);

/********************************/
/* primary function entry point */
/********************************/

int MVEPlayer(const char * filename)
{	
	/*************************/
	/* variable declarations */
	/*************************/

	/* MVE validation tools */
	const char idstring[] = "Interplay MVE File\x1A\0\x1A\0\0\x1\x33\x11";
	const Uint8 idlen = 26;
	char fileTestString[26];

	/* timer vars */
	Uint32 timer_rate = 0, start_time = 0, current_time = 0;
	Uint16 timer_subdivision = 0;
	float ms_per_frame = 0;
	Uint8 frame_timing_adjust = 0;

	/* audio variables */
	SDL_AudioSpec * desired = NULL;
	Uint16 file_audio_flags = 0;
	buffer audio_buffer, audio_data_read, temp_audio_buffer;

	/* video variables */
	Uint16 width_blocks = 0, height_blocks = 0, palette_start = 0;
	SDL_Surface * screen = NULL;
	Uint32 * pixels = NULL, numPalColors = 0, *pal = NULL;
	Uint8 r = 0, g = 0, b = 0;
	buffer map, video;

	/* event vars for determining pause or quit from user */
	SDL_Event event;
	Uint8 keyin = 0;

	/* file handle */
	SDL_RWops * mve = NULL;

	/* data structures to handle file reads */
	chunk ch;
	opcode op;

	/* some counting vars */
	Uint16 i;

	/* initialize audio buffers */
	audio_buffer.data = NULL;
	audio_buffer.length = 0;
	audio_data_read.data = NULL;
	audio_data_read.length = 0;
	temp_audio_buffer.data = NULL;
	temp_audio_buffer.length = 0;

	/*initialize decoding map buffer */
	map.data = NULL;
	map.length = 0;

	/*initialize video data buffer */
	video.data = NULL;
	video.length = 0;

	/*****************************/
	/* end variable declarations */
	/*****************************/


	/******************************************/
	/* validate MVE, init SDL, and begin read */
	/******************************************/
	
	/* initialize file access structure */
	mve = SDL_RWFromFile(filename, "rb");

	/* confirm MVE file open */
	if(!mve) 
		return UNABLE_TO_OPEN_FILE;

	/* read the first 26 bytes */
	if(SDL_RWread(mve, fileTestString, sizeof(char), idlen) < 0) 
		return FILE_TOO_SHORT;

	/* validate MVE idstring */
	if(SDL_memcmp(fileTestString, idstring, idlen) != 0) 
		return NOT_MVE_FILE;
	
	/* open first chunk */
	if(!SDL_RWread(mve, &ch, sizeof(ch), 1)) 
		return CHUNK_READ_FAILED;

	/* read opcode (initial read) */
	if(!SDL_RWread(mve, &op, sizeof(op), 1)) 
		return OPCODE_READ_FAILED;

	/* are opcode parameters within known bounds? */
	if(op.type > 0x15 || op.version > 3) 
		return MVE_CORRUPT;

	/* initialize SDL */
	if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO|SDL_INIT_TIMER) < 0)
	{
		fprintf(stderr, "Could not initialize SDL: %s", SDL_GetError());
		return SDL_INIT_FAILURE;
	}
	atexit(SDL_Quit);

	/***************************************************/
	/* end MVE validation, SDL init, and initial reads */
	/***************************************************/
	
	/*************/
	/* main loop */
	/*************/
	while(!CHKFLAG(QUITTING))
	{
		/********************************************/
		/* event polling for quit / pause from user */
		/********************************************/
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
			case SDL_KEYDOWN:
				keyin = event.key.keysym.sym;
				break;
			case SDL_KEYUP:
				if(event.key.keysym.sym == keyin)
				{
					switch(event.key.keysym.sym)
					{
					case SDLK_q:
					case SDLK_ESCAPE:
					case SDLK_RETURN:
						SETFLAG(QUITTING);
						break;
					case SDLK_SPACE:
						if(CHKFLAG(PAUSED))
							NEGFLAG(PAUSED);
						else
							SETFLAG(PAUSED);
						if(CHKFLAG(PAUSED))
						{
							if(CHKFLAG(AUDIO_PLAYING)) 
								SDL_PauseAudio(1);
							if(CHKFLAG(TIMER_CREATED)) 
								SETFLAG(TIMER_INIT);
						}
						else if(CHKFLAG(AUDIO_PLAYING)) 
							SDL_PauseAudio(0);						
						break;
					default:
						break;
					}
					keyin = 0;
				}
				break;
			case SDL_QUIT:
				SETFLAG(QUITTING);
				break;
			default:
				break;
			}
		}

		if(CHKFLAG(PAUSED) || CHKFLAG(QUITTING)) 
			continue;

		if(CHKFLAG(TIMER_INIT))
		{
			start_time = SDL_GetTicks();
			NEGFLAG(TIMER_INIT);
		}

		/********************/
		/* opcode handling */
		/********************/
    	switch(op.type)
		{
		case STOP_PLAYBACK:

			SETFLAG(QUITTING);

			break;

		case FETCH_NEXT_CHUNK:

			if(!SDL_RWread(mve, &ch, sizeof(ch), 1)) 
				return CHUNK_READ_FAILED;

			break;

		case CREATE_TIMER:

			/* get timer rate (uint32) */
			timer_rate = SDL_ReadLE32(mve);

			/* get timer subdivision (uint16) */
			timer_subdivision = SDL_ReadLE16(mve);
			ms_per_frame = (float)(timer_rate * timer_subdivision / 1000.0);
			SETFLAG(TIMER_CREATED);
			SETFLAG(TIMER_INIT);

			break;

		case INIT_AUDIO_BUFFERS:

			/* init sdl audio settings */
			desired = malloc(sizeof(SDL_AudioSpec));

			/* strip the unknown word out */
			SDL_ReadLE16(mve);

			/* get the file audio flags */
			file_audio_flags = SDL_ReadLE16(mve);

			/* get the sample rate (frequency) */
			desired->freq = SDL_ReadLE16(mve);

			/* fill the known audiospec values */
			desired->samples = 4096;
			desired->callback = MVEPlayerAudioCB;
			desired->userdata = &audio_buffer; /* may have to change that */

			/* fill values gotten from the file audio flags */
			desired->channels = (file_audio_flags & 0x01 ? 2 : 1);
			desired->format = (file_audio_flags & 0x02 ? AUDIO_S16 : AUDIO_U8);
			
			if(desired->channels == 2)
				SETFLAG(AUDIO_STEREO);
			else
				NEGFLAG(AUDIO_STEREO);

			if(desired->format == AUDIO_S16)
				SETFLAG(SAMPLESIZE16);
			else
				NEGFLAG(SAMPLESIZE16);

			/* initialize the audio mode */
			if(SDL_OpenAudio(desired, NULL) < 0)
			{
				fprintf(stderr, "Couldn't open audio: %s", SDL_GetError());
				return COULD_NOT_OPEN_AUDIO;
			}				

			/* set the compressed flag if necessary */
			if(op.version == 1)
				if(file_audio_flags & 0x04 ? 1 : 0)
					SETFLAG(AUDIO_COMPRESSED);
			
			/* we're going to abort if is compressed and (not stereo or not AUDIO_S16) because we don't know what to do */
			/* besides, we've never seen such an MVE before... */
			if(CHKFLAG(AUDIO_COMPRESSED) && (!CHKFLAG(AUDIO_STEREO) || !CHKFLAG(SAMPLESIZE16)))
				return CANNOT_PROCESS_AUDIO;

			/* get rid of rest of opcode data */
			if(op.version == 1)
				SDL_ReadLE32(mve);
			else
				SDL_ReadLE16(mve);

			break;

		case START_AUDIO:

			SETFLAG(AUDIO_PLAYING);
			if(SDL_GetAudioStatus() != SDL_AUDIO_PLAYING)
			{
				SDL_PauseAudio(0);
			}
			break;

		case INIT_VIDEO_BUFFERS:

			/* get the requested dimensions 
			note: this is in 8x8 pixel chunks so multiply by 8 to get requested screen pixels*/
			width_blocks = SDL_ReadLE16(mve);
			height_blocks = SDL_ReadLE16(mve);

			/* ditch the rest of the opcode data (note: this data is where we can get our truecolor flag (16-bit color)*/
			if(op.version > 0)
			{
				SDL_ReadLE16(mve);
				if(op.version > 1)
					SDL_ReadLE16(mve);
			}

			/* if the dimensions are different than the current, set the new ones */
			if(SDL_GetVideoInfo()->current_w != (width_blocks << 3) || SDL_GetVideoInfo()->current_h != (height_blocks << 3))
			{
				/* set SDL_NOFRAME to avoid audio / video desynch during playback due to dragging the screen around */
				screen = SDL_SetVideoMode(width_blocks << 3, height_blocks << 3, 32, SDL_SWSURFACE|SDL_ANYFORMAT);

				/* erase old video buffer data */
				if(old_frame)
				{
					free(old_frame);
					old_frame = NULL;
				}
				if(current_frame)
				{
					free(current_frame);
					current_frame = NULL;
				}

				/* allocate memory for the backbuffers sufficient for the screen pixel buffer */
				old_frame = malloc(screen->h * screen->w * sizeof(Uint32));
				assert(old_frame);
				current_frame = malloc(screen->h * screen->w * sizeof(Uint32));
				assert(current_frame);
			}

			break;

		case SEND_BUFFER_TO_DISPLAY:

			/* if there's time to next scheduled update, force delay to ms_per_frame */
			current_time = SDL_GetTicks();
			if(current_time < start_time + (Uint32)ms_per_frame)
				SDL_Delay(start_time + (Uint32)ms_per_frame - current_time);

			/* send the buffer to the screen */
			SDL_UpdateRect(screen, 0, 0, 0, 0);

			/* strip opcode data; don't know how to catch errors here */
			mve->seek(mve, op.length, SEEK_CUR);
			
			/* increment start time */
			/* note: this fix is for 66+2/3 ms_per_frame; need a general algorithm for figuring this out.  
			   frame_timing_adjust adds 1 to the ms_per_frame n - 1 out of n frames, where n is the right hand side of 
			   frame_timing_adjust %= n ; to get it to do the inverse, remove one ! from !(!frame_timing_adjust) */
			frame_timing_adjust++;
			frame_timing_adjust %= 3;
			start_time += (Uint32)ms_per_frame + !(!frame_timing_adjust);

			break;

		case AUDIO_FRAME:

			/* strip the seq-index and stream-mask */
			SDL_ReadLE16(mve);
			SDL_ReadLE16(mve);

			/* lock SDL out of the audio_buffer; we'll release at the end */
			SDL_LockAudio();

			if(CHKFLAG(AUDIO_COMPRESSED))
			{
				/* send in the raw data via audio_data_read; it will be returned the same way */
				/* read in the compressed data (should be op.length - 4 [the 4 being seq-index and stream-mask]) */
				audio_data_read.length =  op.length - 4;
				audio_data_read.data = malloc(audio_data_read.length);
				assert(audio_data_read.data != NULL);

				if(SDL_RWread(mve, audio_data_read.data, audio_data_read.length, 1) < 0)
					return AUDIO_FRAME_READ_FAILURE;
				MVEPlayerDecodeAudio(&audio_data_read);
			}
			else
			{
				/* get the stream length */
				audio_data_read.length = SDL_ReadLE16(mve);

				/* allocate memory sufficient for the data read */
				audio_data_read.data = malloc(audio_data_read.length);
				assert(audio_data_read.data != NULL);

				/* get audio data */
				if(SDL_RWread(mve, audio_data_read.data, audio_data_read.length, 1) < 0)
					return AUDIO_FRAME_READ_FAILURE;
			}
			/* at this point, we have uncompressed audio in audio_data_read, along with appropriate lengths */

			/* create temp_audio_buffer the size of old audio buffer + new audio data */
			temp_audio_buffer.length = audio_data_read.length + audio_buffer.length;
			temp_audio_buffer.data = malloc(temp_audio_buffer.length);
			assert(temp_audio_buffer.data != NULL);

			/* copy old audio buffer to temp_audio_buffer */
			memcpy(temp_audio_buffer.data, audio_buffer.data, audio_buffer.length);

			/* append new data to temp_audio_buffer */
			memcpy(temp_audio_buffer.data + audio_buffer.length, audio_data_read.data, audio_data_read.length);

			/* free audio buffer */
			if(audio_buffer.data)
				free(audio_buffer.data);

			/* temp_audio_buffer has the requested data, and audio_buffer needs to have it. */
			audio_buffer.data = temp_audio_buffer.data;
			audio_buffer.length = temp_audio_buffer.length;

			/* let SDL have access to the audio_buffer */
			SDL_UnlockAudio();

			/* close off temp_audio_buffer */
			temp_audio_buffer.data = NULL;
			temp_audio_buffer.length = 0;

			/* close off audio_data_read */
			if(audio_data_read.data)
				free(audio_data_read.data);
			audio_data_read.data = NULL;
			audio_data_read.length = 0;

			break;

		case SET_PALETTE:

			/* free previous palette if applicable */
			if(pal)
				free(pal);
			
			/* get # of first entry to fill */
			palette_start = SDL_ReadLE16(mve);
			
			/* get number of entries in palette */
			numPalColors = SDL_ReadLE16(mve);

			/* allocate memory for the palette */
			pal = malloc(sizeof(Uint32) * numPalColors);
			assert(pal != NULL);

			/* fill the palette */
			for(i = palette_start - 1; i < numPalColors; i++)
			{
				SDL_RWread(mve, &r, sizeof(Uint8), 1);
				SDL_RWread(mve, &g, sizeof(Uint8), 1);
				SDL_RWread(mve, &b, sizeof(Uint8), 1);

				/* have to store it as argb */
				/* without the +2, it's too dim */
				pal[i] = (0 << (24 + 2)) | (r << (16 + 2)) | (g << (8 + 2)) | b << (0 + 2);
			}

			break;

		case SET_DECODING_MAP:

			/* kill previous map */
			if(map.data)
				free(map.data);
			
			/* get new map */
			map.data = malloc(sizeof(Uint8) * op.length);
			assert(map.data != NULL);
			map.length = op.length;
			SDL_RWread(mve, map.data, op.length, 1);
			break;

		case VIDEO_DATA:

			/* free the previously allocated video_data buffer */
			if(video.data)
				free(video.data);

			/* allocate enough memory for the video_data buffer */
			video.data = malloc(op.length);
			assert(video.data != NULL);
			video.length = op.length;

			/* read the data */
			SDL_RWread(mve, video.data, op.length, 1);

			/* lock the screen buffer so we can do stuff */
			if(SDL_MUSTLOCK(screen))
				SDL_LockSurface(screen);

			/* decode the video data */
			MVEPlayerDecodeVideo(screen, &video, &map, pal);
			
			/* unlock the screen buffer so SDL can do stuff */
			if(SDL_MUSTLOCK(screen))
				SDL_UnlockSurface(screen);

			break;

		default:

			/* are opcode parameters within known bounds? */
			if(op.type > 0x15 || op.version > 3) 
				return MVE_CORRUPT;

			/* strip opcode data; don't know how to catch errors here */
			mve->seek(mve, op.length, SEEK_CUR);

			break;
		}

		/************************/
		/* end opcode handling */
		/************************/
		
		/* read opcode (trailing read) */
		if(!SDL_RWread(mve, &op, sizeof(op), 1)) 
			return OPCODE_READ_FAILED;

		/* are opcode parameters within known bounds? */
		if(op.type > 0x15 || op.version > 3) 
			return MVE_CORRUPT;

	}
	/***********/
	/* cleanup */
	/***********/

	/* stop audio */
	if(CHKFLAG(AUDIO_PLAYING))
	{
		SDL_PauseAudio(1);
		SDL_CloseAudio();
	}

	/* free audio buffers */
	if(audio_buffer.data)
		free(audio_buffer.data);
	if(audio_data_read.data)
		free(audio_data_read.data);
	if(temp_audio_buffer.data)
		free(temp_audio_buffer.data);

	/* free the audio_spec */
	if(desired)
		free(desired);

	/* free the palette */
	if(pal)
		free(pal);

	/* free the decoding map */
	if(map.data)
		free(map.data);

	/* free the video data */
	if(video.data)
		free(video.data);

	/* free the video buffers */
	if(old_frame)
		free(old_frame);

	if(current_frame)
		free(current_frame);

	/* close the file handle */
	if(mve)
		SDL_RWclose(mve);

	/* seems to have worked. */
	return SUCCESS;
}

/******************************/
/* Audio callback entry point */
/******************************/
void MVEPlayerAudioCB(void *userdata, Uint8 *stream, int len)
{
	Uint8 * temp;
	buffer *audio_buffer = (buffer *)userdata;

	if((Uint32)len > audio_buffer->length)
	{
		/* fprintf(stderr, "SDL wants more data than we have (%i, vs. %i).\n", len, audio_buffer->length); */
		/* give SDL what we have */
		memcpy(stream, audio_buffer->data, audio_buffer->length);
		/* and now we have nothing */
		audio_buffer->length = 0;
	}
	else
	{
		/* copy len bytes to stream */
		memcpy(stream, audio_buffer->data, len);

		/* our buffer length is now whatever we started with minus len */
		audio_buffer->length -= len;

		/* now we need to store the data that wasn't read */
		temp = malloc(audio_buffer->length);
		assert(temp != NULL);

		/* copy from data address + len length bytes 
		(which should be the start of the data that wasn't copied to stream) */
		memcpy(temp, audio_buffer->data + len, audio_buffer->length);

		/* free the memory of the audio_buffer */
		if(audio_buffer->data)
			free(audio_buffer->data);

		/* assign the address of our stored data to the audio_buffer's data */
		audio_buffer->data = temp;
		
		/* close off temp */
		temp = NULL;
	}
}

/*****************************/
/* audio decoder entry point */
/*****************************/
void MVEPlayerDecodeAudio(buffer * in)
{
	/* this only works for stereo dpcm at present */

	Sint16 sample[2];
	Uint16 in_pos = 0, out_pos = 0, channel = 0;
	buffer out;

	/* in->data includes initial stream-len word */
	/* uncompressed streamlen is stream-len bytes */
	out.length = LE16(in->data);
	in_pos +=2;
	out.data = malloc(out.length);
	assert(out.data != NULL);

	/* each byte in the input buffer after the first four (two words: initial L and R values)
	will be expanded to fill a word in the return buffer after decompression. 
	The initial two words will be stored in return buffer as-is. */
	sample[0] = LE16(in->data + in_pos);
	in_pos += 2;
	WRITEWORD((out.data + out_pos), sample[0]);
	out_pos += 2;
	sample[1] = LE16(in->data + in_pos);
	in_pos += 2;
	WRITEWORD((out.data + out_pos), sample[1]);
	out_pos += 2;

	while(in_pos < in->length)
	{
		sample[channel] += interplay_delta_table[*(in->data + in_pos++)];
		WRITEWORD((out.data + out_pos), sample[channel]);
		out_pos += 2;
		channel++;
		channel %= 2;
	}

	/* free the original buffer*/
	if(in->data)
		free(in->data);

	/* assign the incoming buffer the outgoing data for return */
	in->data = out.data;
	in->length = out.length;
}

/**************************************/
/* Video decoding routine entry point */
/**************************************/
void MVEPlayerDecodeVideo(SDL_Surface * screen, buffer * pData, buffer * pMap, Uint32 * pal)
{
#define INCBUF(exp) {exp.data++;exp.length--;}
//#define VIDEODEBUG

	Uint8 hblocks = 0, wblocks = 0, encoding = 0;
	Uint32 screen_buffer_size, *new_frame = NULL, *current = NULL, *old = NULL, *temp = NULL;
	Sint32 x, y, i, j;
	buffer map, data;

	map.data = pMap->data;
	map.length = pMap->length;

	/* there are 14 undocumented bytes at head of video stream */
	data.data = pData->data + 14;
	data.length = pData->length - 14;

	screen_buffer_size = screen->h * screen->w * sizeof(Uint32);

	/* just making sure they were initialized */
	assert(old_frame);
	assert(current_frame);

	/* copy current_frame into old_frame */
	memcpy(old_frame, current_frame, screen_buffer_size);

	/* copy the most recently displayed frame into current_frame */
	memcpy(current_frame, screen->pixels, screen_buffer_size);

	/* copy the old_frame into the new frame */
	memcpy(screen->pixels, old_frame, screen_buffer_size);

#ifdef VIDEODEBUG
	memset(screen->pixels, 0, screen_buffer_size);
#endif

	/* read our number of blocks by dividing width and height by 8 */
	hblocks = screen->h >> 3;
	wblocks = screen->w >> 3;

	// for each row (of 8x8 pixel blocks)
    for(y = 0; y < hblocks; y++) 
	{
		// for each column (of 8x8 pixel blocks)
		for (x = 0; x < wblocks; x++) 
		{
			/* assign the frame buffer pointers to the addresses of the screen 
			buffers so we can do pointer arithmetic on them such as incrementing. */

			new_frame = screen->pixels;
			old = old_frame;
			current = current_frame;

			/* increment pixel pointer past the rendered area on the screen
			y = number of rows, wblocks = number of blocks per row, 8 * 8 = pixel area of the block; 
			y * wblocks * 8 * 8 = all the complete rows of 8x8 pixel blocks rendered.
			x * 8 = upper left pixel of the 8x8 block in this new row to be rendered at this time. */

			new_frame += (y * wblocks * 8 * 8) + (x * 8);
			old += (y * wblocks * 8 * 8) + (x * 8);
			current += (y * wblocks * 8 * 8) + (x * 8);
				
			/* get the decoding map info for this block; low 4 bits if x is even, high 4 bits if x is odd */
			encoding = (*map.data >> (4 * (x & 1))) & 0x0f;

			/* if x is odd, increment the decoding map */
			if(x & 1)
				INCBUF(map);

			/* process the block */
			switch(encoding)
			{
			case 0x0:

				for(j = 0; j < 8; j++)
				{
					for(i = 0; i < 8; i++)
					{
#ifndef VIDEODEBUG
						new_frame[i] = current[i];
#endif
					}
					/* wrap around to next row */
					new_frame += screen->w;
					current += screen->w;
				}
				break;

			case 0x1:

				break;

			case 0x2:
				{
					Uint32 *pixels = new_frame;
					if(*data.data < 56)
					{
						i = 8 + (*data.data % 7);
						j = *data.data / 7;
					}
					else
					{
						i = -14 + ((*data.data - 56) % 29);
						j =   8 + ((*data.data - 56) / 29);
					}
					INCBUF(data);

					pixels += j * screen->w + i;

					for(j = 0; j < 8; j++)
					{
						for(i = 0; i < 8; i++)
						{
#ifndef VIDEODEBUG
							new_frame[i] = pixels[i];
#endif
						}
						pixels += screen->w;
						new_frame += screen->w;
					}
				}
				break;

			case 0x3:

				{
					Uint32 * pixels = new_frame;
					if(*data.data < 56)
					{
						i = -(8 + (*data.data % 7));
						j = -(*data.data / 7);
					}
					else
					{
						i = -(-14 + ((*data.data - 56) % 29));
						j = -(  8 + ((*data.data - 56) / 29));
					}
					INCBUF(data);

					pixels += j * screen->w + i;

					for(j = 0; j < 8; j++)
					{
						for(i = 0; i < 8; i++)
						{
#ifndef VIDEODEBUG
							new_frame[i] = pixels[i];
#endif
						}
						pixels += screen->w;
						new_frame += screen->w;
					}
				}
				break;

			case 0x4:

				i = -8 + (0x0F & *data.data);
				j = -8 + ((0xF0 & *data.data) >> 4);
				INCBUF(data);

				current += j * screen->w + i;

				for(j = 0; j < 8; j++)
				{
					for(i = 0; i < 8; i++)
					{
#ifndef VIDEODEBUG
						new_frame[i] = current[i];
#endif
					}
					current += screen->w;
					new_frame += screen->w;
				}
				break;

			case 0x5:
				{
					Sint8 m, n;

					m = data.data[0];
					n = data.data[1];
					INCBUF(data);
					INCBUF(data);

					current += n * screen->w + m;

					for(j = 0; j < 8; j++)
					{
						for(i = 0; i < 8; i++)
						{
#ifndef VIDEODEBUG
							new_frame[i] = current[i];
#endif
						}
						current += screen->w;
						new_frame += screen->w;
					}
				}
				break;

			case 0x6:

				x += 2;
				if(x >= wblocks)
				{
					y++;
					x %= wblocks;
				}
				break;

			case 0x7:
				{
					Uint8 p[2], b[8], mask;

					p[0] = *data.data;
					INCBUF(data);
					p[1] = *data.data;
					INCBUF(data);

					if(p[0] <= p[1])
					{
						for(i = 0; i < 8; i++)
						{
							b[i] = *data.data;
							INCBUF(data);
						}
						for(j = 0; j < 8; j++)
						{
							for(i = 0, mask = 1; i < 8; i++, mask <<= 1)
							{
#ifndef VIDEODEBUG
								new_frame[i] = pal[p[!!(b[j] & mask)]];
#endif
							}
							new_frame += screen->w;
						}
					}
					else
					{
						b[0] = *data.data;
						INCBUF(data);
						b[1] = *data.data;
						INCBUF(data);

						for(j = 0; j < 4; j++)
						{
							if(!(j % 2))
								mask = 0x01;
							for(i = 0; i < 4; i++, mask <<= 1)
							{
#ifndef VIDEODEBUG
								new_frame[i * 2] = 
									new_frame[i * 2 + 1] =
									new_frame[i * 2 + screen->w] = 
									new_frame[i * 2 + 1 + screen->w] = pal[p[!!(b[j / 2] & mask)]];
#endif
							}
							new_frame += 2 * screen->w;
						}
					}
				} /* end case block */
				break;

			case 0x8:
				{
					Uint8 p[8], b[8], mask, k;
					if(data.data[0] <= data.data[1])
					{
						for(j = 0; j < 4; j++)
						{
							for(i = 0; i < 2; i++)
							{
								p[i + j * 2] = *data.data;
								INCBUF(data);
							}
							for(i = 0; i < 2; i++)
							{
								b[i + j * 2] = *data.data;
								INCBUF(data);
							}
						}

						for(k = 0; k < 4; k++)
						{
							if(k == 2)
							{
								new_frame -= screen->w * 8;
								new_frame += 4;
							}
							for(j = 0; j < 4; j++)
							{
								if(!(j & 1))
									mask = 1;
								for(i = 0; i < 4; i++)
								{
#ifndef VIDEODEBUG
									new_frame[i] = pal[p[!!(b[j / 2 + 2 * k] & mask) + 2 * k]];
#endif
									mask <<= 1;
								} /* for i */
								new_frame += screen->w;
							} /* for j */
						} /* for k */
					} /* if p0 <= p1 */
					else
					{
						for(j = 0; j < 2; j++)
						{
							for(i = 0; i < 2; i++)
							{
								p[j * 2 + i] = *data.data;
								INCBUF(data);
							}
							for(i = 0; i < 4; i++)
							{
								b[j * 4 + i] = *data.data;
								INCBUF(data);
							}
						}
						if(p[2] > p[3])
						{
							/* horizontally split block */
							for(j = 0; j < 8; j++)
							{
								for(i = 0, mask = 1; i < 8; i++, mask <<= 1)
								{
#ifndef VIDEODEBUG
									new_frame[i] = pal[p[!!(b[j] & mask) + (j > 3 ? 2 : 0)]];
#endif
								}
								new_frame += screen->w;
							}
						}
						else
						{
							/* vertically split block */
							for(k = 0; k < 2; k++)
							{
								if(k == 1)
								{
									new_frame -= screen->w * 8;
									new_frame += 4;
								}
								for(j = 0; j < 4; j++)
								{
									mask = 1;
									for(i = 0; i < 8; i++)
									{
#ifndef VIDEODEBUG
										new_frame[i % 4] = pal[p[!!(b[j + k * 4] & mask) + k * 2]];
#endif
										mask <<= 1;
										if(i == 3)
											new_frame += screen->w;
									}
									new_frame += screen->w;
								}
							}
						}
					} /* end if/else block */					
				} /* end case block */

				break;

			case 0x9:
				{
					Uint8 p[4], b[16];
					for(i = 0; i < 4; i++)
					{
						p[i] = data.data[0];
						INCBUF(data);
					}
					if(p[0] <= p[1] && p[2] <= p[3])
					{
						Uint16 mask;
						for(i = 0; i < 16; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
						for(j = 0; j < 8; j++)
						{
							for(i = 0, mask = 3; i < 8; i++, mask <<= 2)
							{
#ifndef VIDEODEBUG
								new_frame[i] = pal[p[(mask & ((b[2 * j + 1] << 8) | b[2 * j])) >> (2 * i)]];
#endif
							}
							new_frame += screen->w;
						}
					}
					if(p[0] <= p[1] && p[2] > p[3])
					{
						Uint8 mask;
						for(i = 0; i < 4; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
						for(j = 0; j < 8; j++)
						{
							mask = 3;
							for(i = 0; i < 8; i++)
							{
#ifndef VIDEODEBUG
								new_frame[i] = pal[p[(b[j / 2] & mask) >> 2 * (i / 2)]];
#endif
								if(i & 1)
									mask <<= 2;
							}
							new_frame += screen->w;
						}
					}
					if(p[0] > p[1] && p[2] <= p[3])
					{
						Uint8 mask;
						for(i = 0; i < 8; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
						for(j = 0; j < 8; j++)
						{
							mask = 3;
							for(i = 0; i < 8; i++)
							{
#ifndef VIDEODEBUG
								new_frame[i] = pal[p[(b[j] & mask) >> 2 * (i / 2)]];
#endif
								if(i & 1)
									mask <<= 2;
							}
							new_frame += screen->w;
						}
					}
					if(p[0] > p[1] && p[2] > p[3])
					{
						Uint16 mask;
						for(i = 0; i < 8; i++)
						{
							b[i] = data.data[0];
							INCBUF(data);
						}
						for(j = 0; j < 4; j++)
						{
							mask = 3;
							for(i = 0; i < 8; i++)
							{
#ifndef VIDEODEBUG
								new_frame[i] =
									new_frame[i + screen->w] = pal[p[((b[j * 2 + 1] << 8 | b[j * 2]) & mask) >> i * 2]];
#endif
								mask <<= 2;
							}
							new_frame += screen->w * 2;
						}
					}
				}
				break;

			case 0xA:
				{
					Uint8 p[16], b[16], mask, k;
					if(data.data[0] <= data.data[1])
					{
						for(i = 0; i < 4; i++)
						{
							for(j = 0; j < 4; j++)
							{
								p[i * 4 + j] = *data.data;
								INCBUF(data);
							}
							for(j = 0; j < 4; j++)
							{
								b[i * 4 + j] = *data.data;
								INCBUF(data);
							}
						}
						for(k = 0; k < 4; k++)
						{
							for(j = 0; j < 4; j++)
							{
								for(i = 0, mask = 3; i < 4; i++, mask <<= 2)
								{
#ifndef VIDEODEBUG
									new_frame[i] = pal[p[((b[k * 4 + j] & mask) >> (i * 2)) + k * 4]];
#endif
								}
								new_frame += screen->w;
							}
							if(k == 1)
							{
								new_frame -= 8 * screen->w;
								new_frame += 4;
							}
						}
					}
					else
					{
						for(i = 0; i < 2; i++)
						{
							for(j = 0; j < 4; j++)
							{
								p[i * 4 + j] = *data.data;
								INCBUF(data);
							}
							for(j = 0; j < 8; j++)
							{
								b[i * 8 + j] = *data.data;
								INCBUF(data);
							}
						}
						if(p[4] <= p[5])
						{
							for(k = 0; k < 2; k++)
							{
								for(j = 0; j < 8; j++)
								{
									for(i = 0, mask = 3; i < 4; i++, mask <<= 2)
									{
#ifndef VIDEODEBUG
										new_frame[i] = pal[p[((b[j + 8 * k] & mask) >> (i * 2)) + k * 4]];
#endif
									}
									new_frame += screen->w;
								}
								if(k == 0)
								{
									new_frame -= 8 * screen->w;
									new_frame += 4;
								}
							}
						}
						else
						{
							for(k = 0; k < 2; k++)
							{
								for(j = 0; j < 4; j++)
								{
									for(i = 0, mask = 3; i < 8; i++, mask <<= 2)
									{
										if(i == 4)
											mask = 3;
#ifndef VIDEODEBUG
										new_frame[i] = pal[p[((b[2 * j + (i > 3 ? 1 : 0) + k * 8] & mask) >> ((i % 4) * 2)) + k * 4]];
#endif
									}
									new_frame += screen->w;
								}
							}
						}
					}
				}
				break;

			case 0xB:

				for(j = 0; j < 8; j++)
				{
					for(i = 0; i < 8; i++)
					{
#ifndef VIDEODEBUG
						new_frame[i] = pal[*data.data];
#endif
						INCBUF(data);
					}
					new_frame += screen->w;
				}
				break;

			case 0xC:
				{
					Uint8 p[16];
					for(i = 0; i < 16; i++)
					{
						p[i] = data.data[0];
						INCBUF(data);
					}
					for(j = 0; j < 8; j++)
					{
						for(i = 0; i < 8; i++)
						{
#ifndef VIDEODEBUG
							new_frame[i] = pal[p[i / 2 + (j / 2) * 4]];
#endif
						}
						new_frame += screen->w;
					}
				}
				break;

			case 0xD:
				{
					Uint8 p[4];
					for(i = 0; i < 4; i++)
					{
						p[i] = data.data[0];
						INCBUF(data);
					}
					for(j = 0; j < 8; j++)
					{
						for(i = 0; i < 8; i++)
						{
#ifndef VIDEODEBUG
							new_frame[i] = pal[p[2 * (j > 4 ? 1 : 0) + (i > 4 ? 1 : 0)]];
#endif
						}
						new_frame += screen->w;
					}
				}
				break;

			case 0xE:

				for(j = 0; j < 8; j++) 
				{
					for(i = 0; i < 8; i++)
#ifndef VIDEODEBUG
						new_frame[i] = pal[data.data[0]];
#endif
					new_frame += screen->w;
				}
				INCBUF(data);
				break;

			case 0xF:

				for(j = 0; j < 8; j++)
				{
					for(i = 0; i < 8; i++)
#ifndef VIDEODEBUG
						new_frame[i] = pal[data.data[((i + j) & 1)]];
#endif
					new_frame += screen->w;
				}
				INCBUF(data);
				INCBUF(data);
				break;
			} /* end switch(encoding) */
		} /* end for(x) */
	} /* end for(y) */
}